import asyncio
from typing import Any
from typing import AsyncIterator
from typing import Dict
from typing import List
from typing import Tuple

__func_alias__ = {"iter_": "iter", "publishers_": "publishers"}
STOP_ITERATION = object()


def __init__(hub):
    hub.ingress.RUN_FOREVER = True
    hub.ingress.QUEUE = None


async def queue(hub):
    if hub.ingress.QUEUE is None:
        hub.ingress.QUEUE = asyncio.Queue()
    return hub.ingress.QUEUE


async def publish(hub, event: Dict):
    # initialize the queue with the current loop
    await hub.ingress.init.queue()
    await hub.ingress.QUEUE.put(event)


async def iter_(hub) -> AsyncIterator:
    """
    Iterate over the ingress queue until RUN_FOREVER is unset
    """
    await hub.ingress.init.queue()
    while hub.ingress.RUN_FOREVER:
        event = await hub.ingress.QUEUE.get()
        if event is STOP_ITERATION:
            hub.log.debug("Cancelled iteration of ingress queue")
            return
        yield event


async def publishers_(
    hub, contexts: Dict[str, Any]
) -> List[Tuple["pop.contract.ContractedAsync", Dict[str, Any]]]:
    """
    For all the contexts and ingress plugins, find the
    """
    hub.log.debug("Collecting publisher plugins")
    ret = []

    for plugin in hub.ingress:
        name = plugin.__name__

        # If the listen function of ingress has ctx, then populate it from acct
        if "ctx" in plugin.publish.signature.parameters:
            # Iterate over the defined profiles for the plugin
            for profile, ctx in contexts.get(name, {}).items():
                # Inject the ctx from the profile into the listener
                hub.log.debug(f"Created {name} ingress with ctx")
                ret.append((plugin.publish, ctx))
        else:
            hub.log.debug(f"Created {name} ingress")
            ret.append((plugin.publish, None))

    return ret


async def vomit(
    hub,
    event: Dict,
    publishers: List[Tuple["pop.contract.ContractedAsync", Dict[str, Any]]],
):
    """
    Post an event to all the event queues
    """
    coros = []
    for publisher, ctx in publishers:
        if ctx is None:
            coros.append(publisher(event))
        else:
            coros.append(publisher(ctx, event))

    for coro in asyncio.as_completed(coros):
        await coro



CLI_CONFIG = {
    "ingress_profiles": {"nargs": "*", "group": "evbus"},
    "random": {"action": "store_true", "group": "evbus"},
    # acct options
    "acct_file": {"source": "acct", "os": "ACCT_FILE"},
    "acct_key": {"source": "acct", "os": "ACCT_KEY"},
    # rend options
    "output": {"source": "rend"},
}
CONFIG = {
    "ingress_profiles": {
        "help": "The acct profile names to allowlist.  If none are specified then all profiles will be used",
        "default": [],
    },
    "random": {
        "help": "Run the random number generator that automatically populates the event bus",
        "default": False,
        "type": bool,
    },
}
SUBCOMMANDS = {}
DYNE = {
    "acct": ["acct"],
    "log": ["log"],
    "evbus": ["evbus"],
    "ingress": ["ingress"],
}



import asyncio
import logging.handlers
import queue


def wrapper(hub, **kwargs):
    class AsyncQueueWrapper:
        def __init__(self):
            self.wait_queue = queue.Queue(**kwargs)

        @property
        def _loop(self) -> asyncio.AbstractEventLoop:
            return hub.pop.loop.CURRENT_LOOP

        @property
        def _queue(self) -> asyncio.Queue:
            return hub.ingress.QUEUE

        def put_nowait(self, record):
            if self._queue is None:
                # The asynchronous ingress queue has not yet been initialized, put it on the waiter
                self.wait_queue.put(record)
                return
            else:
                # If we got this far, clear the waiter queue onto the real one
                while not self.wait_queue.empty():
                    self.put_nowait(self.wait_queue.get())

            # Put new records onto the queue
            if self._loop is None:
                # No loop but we have a queue? put it on the queue immediately
                self._queue.put_nowait(record)
            else:
                # There is a loop and a queue, create a task
                coro = self._queue.put(record)
                self._loop.create_task(coro)

    return AsyncQueueWrapper()


def setup(hub, conf):
    """
    Log to the ingress queue
    """
    root = logging.getLogger()

    raw_level = conf["log_level"].strip().lower()
    if raw_level.isdigit():
        hub.log.INT_LEVEL = int(raw_level)
    else:
        hub.log.INT_LEVEL = hub.log.LEVEL.get(raw_level, root.level)

    root.setLevel(hub.log.INT_LEVEL)
    cf = logging.Formatter(fmt=conf["log_fmt_console"], datefmt=conf["log_datefmt"])
    ch = logging.StreamHandler()
    ch.setLevel(hub.log.INT_LEVEL)
    ch.setFormatter(cf)
    root.addHandler(ch)

    ff = logging.Formatter(fmt=conf["log_fmt_logfile"], datefmt=conf["log_datefmt"])
    _, kwargs = hub.render.cli.args(conf["log_handler_options"])

    qh = logging.handlers.QueueHandler(hub.log.queue.wrapper(**kwargs))
    qh.setLevel(hub.log.INT_LEVEL)
    qh.setFormatter(ff)
    root.addHandler(qh)



import asyncio
from typing import Any
from typing import Dict
from typing import List

from dict_tools.data import NamespaceDict

__func_alias__ = {"print_": "print"}

STOP_ITERATION = object()


def __init__(hub):
    hub.evbus.STARTED = False
    hub.evbus.RUN_FOREVER = True
    hub.evbus.BUS = None
    for dyne in ("acct", "ingress", "output"):
        hub.pop.sub.add(dyne_name=dyne)


async def queue(hub):
    """
    Initialize the event bus queue in the current loop
    """
    if hub.evbus.BUS is None:
        hub.evbus.BUS = asyncio.Queue()
    return hub.evbus.BUS


def cli(hub):
    hub.pop.config.load(["evbus", "acct", "rend"], cli="evbus")
    hub.pop.loop.create()
    hub.pop.Loop.run_until_complete(hub.evbus.init.cli_apply())


async def cli_apply(hub):
    sub_profiles = await hub.evbus.init.profiles(
        subs=["evbus"],
        acct_file=hub.OPT.acct.acct_file,
        acct_key=hub.OPT.acct.acct_key,
        acct_profiles=hub.OPT.evbus.ingress_profiles,
    )

    coros = [
        hub.pop.Loop.create_task(
            hub.evbus.init.start(sub_profiles=sub_profiles.get("evbus", {}))
        ),
        hub.evbus.init.print(renderer=hub.OPT.rend.output or "json"),
    ]

    if hub.OPT.evbus.random:
        coros.append(hub.evbus.test.start())

    try:
        for task in asyncio.as_completed(coros):
            await task
    except KeyboardInterrupt:
        hub.log.debug("Shutting down Event Bus")


async def start(hub, sub_profiles: Dict[str, Dict[str, Any]]):
    hub.log.debug("Starting event bus")
    publishers = await hub.ingress.init.publishers(sub_profiles)
    hub.evbus.STARTED = True

    # Initialize the queue in the current loop
    await hub.evbus.init.queue()

    while hub.evbus.RUN_FOREVER:
        # Block until another event is added to the queue
        await hub.evbus.BUS.join()
        event = await hub.evbus.BUS.get()
        if event is STOP_ITERATION:
            hub.log.debug("Event bus was forced to stop")
            return
        await hub.ingress.init.vomit(event, publishers)
        hub.evbus.BUS.task_done()


async def stop(hub):
    """
    Stop the event bus
    """
    if hub.evbus.BUS is not None:
        hub.log.debug("Stopping the event bus")
        await hub.evbus.BUS.put(STOP_ITERATION)

    hub.evbus.RUN_FOREVER = False


async def join(hub):
    while not hub.evbus.STARTED:
        await asyncio.sleep(0)


async def print_(hub, renderer: str = "json"):
    """
    Just print out the queue forever
    """
    await hub.ingress.init.queue()
    hub.log.debug("Started evbus print loop")
    while hub.evbus.RUN_FOREVER:
        await asyncio.sleep(0)
        event = await hub.ingress.QUEUE.get()
        if event is hub.ingress.init.STOP_ITERATION:
            return

        print(hub.output[renderer].display(event))
        print("-" * 80)


async def profiles(
    hub,
    subs: List[str],
    acct_file: str = None,
    acct_key: str = None,
    acct_profiles: List[str] = None,
) -> Dict[str, Any]:
    """
    Read the acct information from the named subs and return the context
    """
    if acct_file and acct_key:
        hub.log.debug("Reading profiles from acct")
        all_profiles = await hub.acct.init.profiles(acct_file, acct_key)
    else:
        all_profiles = {}

    ret_profiles = NamespaceDict()

    # If acct_profiles is empty then use all profiles
    if acct_profiles:
        for provider, profile in all_profiles.items():
            ret_profiles[provider] = {}
            for name, info in profile.items():
                # Remove profiles that aren't specified in the acct_profiles
                if name in acct_profiles:
                    ret_profiles[provider][name] = info

    # Create sub profiles
    new_profiles = await hub.acct.init.process(subs, ret_profiles)

    sub_profiles = {}
    for sub in subs:
        sub_profiles.update(new_profiles.get(sub, {}))

    return sub_profiles
